#include "StdAfx.h"
#include "DropTarget.h"
#include <shellapi.h>
#include "misc.h"


CDropTarget::CDropTarget(HWND hwnd, ISVALIDFILECALLBACK callback)
{
	m_cRef = 1;
	m_hWnd = hwnd;
	IsValidCallback = callback;
}

CDropTarget::~CDropTarget()
{
}

STDMETHODIMP CDropTarget::QueryInterface(REFIID riid, void **ppvObject)
{
	*ppvObject = NULL;

	if (IsEqualIID(riid, IID_IUnknown) || IsEqualIID(riid, IID_IDropTarget))
		*ppvObject = static_cast<IDropTarget *>(this);
	else
		return E_NOINTERFACE;

	AddRef();

	return S_OK;
}

STDMETHODIMP_(ULONG) CDropTarget::AddRef()
{
	return InterlockedIncrement(&m_cRef);
}

STDMETHODIMP_(ULONG) CDropTarget::Release()
{
	if (InterlockedDecrement(&m_cRef) == 0) {
		delete this;
		return 0;
	}

	return m_cRef;
}

STDMETHODIMP CDropTarget::DragEnter(IDataObject *pDataObj, DWORD grfKeyState, POINTL pt, DWORD *pdwEffect)
{
	HRESULT   hr;
	FORMATETC formatetc;
	STGMEDIUM   medium;
	UINT uiSize;
	TCHAR szPath[_MAX_PATH];

	m_bSupportFormat = FALSE;
	*pdwEffect = DROPEFFECT_NONE;

	formatetc.cfFormat = CF_HDROP;
	formatetc.ptd      = NULL;
	formatetc.dwAspect = DVASPECT_CONTENT;
	formatetc.lindex   = -1;
	formatetc.tymed    = TYMED_HGLOBAL;

	hr = pDataObj->QueryGetData(&formatetc);
	if (hr != S_OK) {
		return hr;
	}
	else {
		pDataObj->GetData( &formatetc, &medium );

		uiSize = DragQueryFile( (HDROP)medium.hGlobal, 0xFFFFFFFF, NULL, 0 );

		for ( UINT c = 0; c < uiSize; c++ ) {

			DragQueryFile( (HDROP)medium.hGlobal, c, szPath, sizeof(szPath) );

			if ( ( IsValidCallback ) && IsValidCallback( szPath ) ) {
				m_bSupportFormat = TRUE;
				*pdwEffect = DROPEFFECT_COPY;
			}
		}
	}

	return S_OK;
}

STDMETHODIMP CDropTarget::DragOver(DWORD grfKeyState, POINTL pt, DWORD *pdwEffect)
{
	*pdwEffect = DROPEFFECT_COPY;

	if (!m_bSupportFormat)
		*pdwEffect = DROPEFFECT_NONE;

	return S_OK;
}

STDMETHODIMP CDropTarget::DragLeave()
{
	return S_OK;
}

STDMETHODIMP CDropTarget::Drop(IDataObject *pDataObj, DWORD grfKeyState, POINTL pt, DWORD *pdwEffect)
{
	HRESULT   hr;
	FORMATETC formatetc;
	STGMEDIUM medium;

	if ( m_bSupportFormat ) {
		*pdwEffect = DROPEFFECT_COPY;

		formatetc.cfFormat = CF_HDROP;
		formatetc.ptd      = NULL;
		formatetc.dwAspect = DVASPECT_CONTENT;
		formatetc.lindex   = -1;
		formatetc.tymed    = TYMED_HGLOBAL;

		hr = pDataObj->GetData(&formatetc, &medium);
		if (FAILED(hr)) {
			*pdwEffect = DROPEFFECT_NONE;
			return E_FAIL;
		}

		PostMessage( m_hWnd, WM_DROPFILES, (WPARAM)medium.hGlobal, 0 );
	}
	else {
		*pdwEffect = DROPEFFECT_NONE;
		return E_FAIL;
	}

	return S_OK;
}
