#include "StdAfx.h"
#include "MainDialog.h"
#include "resource.h"
#include "misc.h"
#include <PowrProf.h>


#define IDM_SYS_ABOUT 0x10

const COLORREF g_clrText = ::GetSysColor(COLOR_WINDOWTEXT);
const COLORREF g_clrBkgnd = ::GetSysColor(COLOR_WINDOW);
const COLORREF g_clrHText = ::GetSysColor(COLOR_HIGHLIGHTTEXT);
const COLORREF g_clrHBkgnd = ::GetSysColor(COLOR_HIGHLIGHT);
const COLORREF g_clrPercent = ::GetSysColor(COLOR_HOTLIGHT);
const COLORREF g_clrHPercent = ::GetSysColor(COLOR_BTNFACE);

MainDialog::MainDialog(UINT idResource)
	: ReposDialog( idResource )
	, pDropTarget( NULL )
	, hList( NULL )
	, nCurItem( 0 )
	, bInDemux( FALSE )
	, m_aboutDlg( NULL )
{
	static REPOS_ITEM items[] = {
		{ IDC_LIST,				ANCHOR_LEFT | ANCHOR_TOP | ANCHOR_RIGHT | ANCHOR_BOTTOM },
		{ IDC_BTN_START,		ANCHOR_LEFT | ANCHOR_BOTTOM },
		{ IDOK,					ANCHOR_RIGHT | ANCHOR_BOTTOM },
		{ IDC_CHECK_SUSPEND,	ANCHOR_LEFT | ANCHOR_BOTTOM },
		{ IDC_CHECK_ERRORSPLIT,	ANCHOR_LEFT | ANCHOR_BOTTOM },
	};

	CoInitialize(NULL);
	m_ptrTaskbarList3.CoCreateInstance(CLSID_TaskbarList);
	reposItems = items;
	reposNum = _countof(items);
}

MainDialog::~MainDialog(void)
{
	UnregisterClass( szWindowClass, hInst );
}

INT_PTR CALLBACK MainDialog::DlgProc( UINT uMsg, WPARAM wParam, LPARAM lParam )
{
	ReposDialog::ReposProc( uMsg, wParam, lParam );

	switch( uMsg )
	{
	HANDLE_MSG( m_hDlg, WM_INITDIALOG, Cls_OnInitDialog );
	HANDLE_MSG( m_hDlg, WM_DROPFILES, Cls_OnDropFiles );
	HANDLE_MSG( m_hDlg, WM_SYSCOMMAND, Cls_OnSysCommand );
	HANDLE_MSG( m_hDlg, WM_COMMAND, Cls_OnCommand );
	HANDLE_MSG( m_hDlg, WM_DESTROY, Cls_OnDestroy );

	case WM_GETMINMAXINFO:	MainSetMin( &((MINMAXINFO*)lParam)->ptMinTrackSize ); break;
	case WM_NOTIFY:			return Cls_OnNotify( wParam, lParam );
	case WM_DEMUXPROGRESS:	Cls_OnDemuxProgress( wParam ); break;
	case WM_DEMUXSPLIT:		Cls_OnDemuxSplit( wParam ); break;

	case WM_DEMUXFINISHED:
	case WM_DEMUXABORTED:
		Next( uMsg == WM_DEMUXFINISHED ? STATE_FINISHED : STATE_FAILED );
		::SetFocus( GetDlgItem( m_hDlg, IDOK ) );
		break;

	default:
		return DefWindowProc(m_hDlg, uMsg, wParam, lParam);
		break;
	}

	return 0;
}

BOOL MainDialog::InitInstance( HWND hParent )
{
	WNDCLASSEX wcex;

	wcex.cbSize = sizeof(WNDCLASSEX);

	wcex.style			= CS_HREDRAW | CS_VREDRAW;
	wcex.lpfnWndProc	= (WNDPROC)_RootWndProc;
	wcex.cbClsExtra		= 0;
	wcex.cbWndExtra		= DLGWINDOWEXTRA;
	wcex.hInstance		= hInst;
	wcex.hIcon			= LoadIcon(hInst, MAKEINTRESOURCE(IDI_DIALOG));
	wcex.hCursor		= LoadCursor(NULL, IDC_ARROW);
	wcex.hbrBackground	= (HBRUSH)(COLOR_BTNFACE+1);
	wcex.lpszMenuName	= NULL;
	wcex.lpszClassName	= szWindowClass;
	wcex.hIconSm		= NULL;

	return RegisterClassEx(&wcex);
}

BOOL MainDialog::DoModeless(HWND hParent, int nCmdShow)
{
	BOOL bRet = FALSE;

	if ( InitInstance() ) {
		if ( bRet = Dialog::DoModeless( hParent, nCmdShow ) ) {
			SetWindowText( m_hDlg, szTitle );
		}
	}
	return bRet;
}

void MainDialog::InsertSystemMenu( void )
{
	MENUITEMINFO mii;
	HMENU hSysMenu;
	TCHAR szTemp[256];
	int nMenuItemCount;

	memset ( &mii, 0, sizeof(MENUITEMINFO) );
	mii.cbSize = sizeof(MENUITEMINFO);

	mii.fMask = MIIM_TYPE;
	mii.fType = MFT_SEPARATOR;

	// Zp[^ǉ
	hSysMenu = GetSystemMenu( m_hDlg, FALSE );
	nMenuItemCount = GetMenuItemCount ( hSysMenu );
	InsertMenuItem ( hSysMenu, nMenuItemCount -1, TRUE, &mii );

	LoadString( hInst, IDS_ABOUTBOX, szTemp, _countof(szTemp) );
	mii.fMask	   = MIIM_ID | MIIM_TYPE;
	mii.fType	   = MFT_STRING;
	mii.wID 	   = IDM_SYS_ABOUT;
	mii.dwTypeData = szTemp;
	mii.cch 	   = _tcslen(szTemp);

	InsertMenuItem( hSysMenu, nMenuItemCount -1, TRUE, &mii );
	DrawMenuBar( m_hDlg );
}

int MainDialog::InitControls(void)
{
	LVCOLUMN	lvc;
	int 		i;
	const UINT	captionTable[] = { 0, IDS_COLUMN_PATH, IDS_COLUMN_STATUS };
	int			width[] = { 40, 300, 100 };
	const int	clmNum = _countof(captionTable);
	int 		err = 0;

	imageList = ImageList_Create( 16, 16, ILC_COLOR8 | ILC_MASK, 2, 2 );
	ImageList_AddIcon( imageList, LoadIcon( hInst, MAKEINTRESOURCE( IDI_ALERT) ) );
	ImageList_AddIcon( imageList, LoadIcon( hInst, MAKEINTRESOURCE( IDI_FINISH ) ) );
	ListView_SetImageList( hList, imageList, LVSIL_SMALL );

	lvc.mask = LVCF_WIDTH | LVCF_TEXT | LVCF_SUBITEM;	 // LtO
	for (i = 0; i < clmNum; i++)
	{
		TCHAR szCaption[256];
		if ( captionTable[i] != 0 ) {
			LoadString( hInst, captionTable[i], szCaption, 256 );
		}
		else {
			szCaption[0] = 0;
		}

		lvc.iSubItem	= i;					// TuACeԍ
		lvc.pszText 	= szCaption;			// oeLXg
		lvc.cx			= width[i];				// 
		if ( ListView_InsertColumn( hList, i, &lvc ) == -1 ) {
			err = 1;
			break;
		}
	}

	DWORD dwStyle = ListView_GetExtendedListViewStyle( hList );
	dwStyle |= LVS_EX_SUBITEMIMAGES | LVS_EX_FULLROWSELECT;
	ListView_SetExtendedListViewStyle( hList, dwStyle );

	EnableItem( IDC_BTN_START, FALSE );

	return err;
}

int MainDialog::GetImageIndex( PROG_STATE state )
{
	int idx;

	switch( state )
	{
	case STATE_FAILED:
		idx = 0;
		break;
	case STATE_FINISHED:
		idx = 1;
		break;
	default:
		idx = -1;
	}

	return idx;
}

int MainDialog::ListInsertItem( TCHAR *path )
{
	LVITEM		lvi;
	int 		index = 0;
	int 		err = 0;
	int			num;

	num = ListView_GetItemCount( hList );

	// ACR
	lvi.mask = LVIF_IMAGE;
	lvi.iItem = num;
	lvi.iSubItem = 0;
	lvi.iImage = GetImageIndex( STATE_NONE );
	if ((index = ListView_InsertItem( hList, &lvi)) == -1) err = 1;

	// V{bNpX
	if (!err) {
		lvi.mask = LVIF_TEXT;
		lvi.iItem = num;
		lvi.iSubItem = 1;
		lvi.pszText = path;
		if ( !ListView_SetItem( hList, &lvi ) ) err = 1;
	}
	// O
	if (!err) {
		lvi.iItem = num;
		lvi.iSubItem = 2;
		lvi.pszText = _T("ҋ@");
		if ( !ListView_SetItem( hList, &lvi ) ) err = 1;
	}

	return err;
}

INT_PTR MainDialog::ListCustom( LPNMLVCUSTOMDRAW lplvcd )
{
	TCHAR text[256];
	RECT rcItem;

	switch( lplvcd->nmcd.dwDrawStage )
	{
	case CDDS_PREPAINT:
		return CDRF_NOTIFYSUBITEMDRAW;

	case CDDS_ITEMPREPAINT:
		return CDRF_NOTIFYSUBITEMDRAW;

	case CDDS_ITEMPREPAINT | CDDS_SUBITEM:
		int iCol = lplvcd->iSubItem;
		int iRow = lplvcd->nmcd.dwItemSpec;
		ListView_GetItemText( hList, iRow, iCol, text, sizeof(text) );
		GetCellRect( iRow, iCol, LVIR_BOUNDS, &rcItem );

		HDC hdc = lplvcd->nmcd.hdc;
		int State = ListView_GetItemState( hList, iRow, LVIS_CUT|LVIS_SELECTED|LVIS_FOCUSED|LVIS_STATEIMAGEMASK);
		FillSolidRect( hdc, &rcItem, ( State &LVIS_SELECTED ) ? g_clrHBkgnd : g_clrBkgnd );
		InflateRect( &rcItem, -2, -1 );

		if ( text[_tcslen(text)-1] == _T('%') ) {
			Rectangle( hdc, rcItem.left, rcItem.top, rcItem.right, rcItem.bottom );
			InflateRect( &rcItem, -1, -1 );
			RECT rcLeft, rcRight;
			rcLeft = rcRight = rcItem;
			rcLeft.right = rcLeft.left + MulDiv( _ttoi(text), rcItem.right - rcItem.left, 100 );
			rcRight.left = rcLeft.right;
			if ( ( State & LVIS_SELECTED ) ) {
				FillSolidRect( hdc, &rcLeft, g_clrHPercent );
				FillSolidRect( hdc, &rcRight, g_clrBkgnd );
				SetTextColor( hdc, g_clrText );
				DrawText( hdc, text, _tcslen(text), &rcItem, DT_VCENTER | DT_CENTER | DT_SINGLELINE );
			}
			else {
				FillSolidRect( hdc, &rcLeft, g_clrPercent);
				FillSolidRect( hdc, &rcRight, g_clrHPercent);
				HRGN hrgn;
				hrgn = CreateRectRgnIndirect( &rcLeft );
				SelectClipRgn( hdc, hrgn );
				SetTextColor( hdc, g_clrBkgnd );
				DrawText( hdc, text, _tcslen(text), &rcItem, DT_VCENTER | DT_CENTER | DT_SINGLELINE );
				SetRectRgn( hrgn, rcRight.left, rcRight.top, rcRight.right, rcRight.bottom );
				SelectClipRgn( hdc, hrgn );
				SetTextColor( hdc, g_clrText );
				DrawText( hdc, text, _tcslen(text), &rcItem, DT_VCENTER | DT_CENTER | DT_SINGLELINE );
				SelectClipRgn( hdc, NULL );
				DeleteObject( hrgn );
			}
		}
		else {
			LVITEM lvi;
			lvi.iSubItem = iCol;
			lvi.iItem = iRow;
			lvi.mask = LVIF_IMAGE;
			if ( ListView_GetItem( hList, &lvi ) && ( lvi.iImage != -1 ) ) {
				return CDRF_DODEFAULT;
			}
			SetTextColor( hdc, (State & LVIS_SELECTED) ? g_clrHText:g_clrText );
			if ( _tcsncmp( text + 1, _T(":\\") , 2 ) == 0 ) {
				PathCompactPath( hdc, text, rcItem.right - rcItem.left );
				DrawText( hdc, text, _tcslen(text), &rcItem, DT_NOCLIP | DT_LEFT | DT_VCENTER );
			}
			else {
				DrawText( hdc, text, _tcslen(text), &rcItem, DT_NOCLIP | DT_LEFT | DT_VCENTER | DT_END_ELLIPSIS );
			}
		}
		return CDRF_SKIPDEFAULT;
	}

	return 0;
}

BOOL MainDialog::GetCellRect(int iRow, int iCol, int nArea, RECT *rect)
{
	if ( iCol ) {
		return ListView_GetSubItemRect( hList, iRow, iCol, nArea, rect );
	}

	if ( ListView_GetItemCount( hList ) == 1 ) {
		return ListView_GetItemRect( hList, iRow, rect, nArea);
	}

	RECT rCol1;
	if ( !ListView_GetSubItemRect( hList, iRow, 1, nArea, &rCol1 ) ) {
		return FALSE;
	}

	if ( !ListView_GetItemRect( hList, iRow, rect, nArea ) ) {
		return FALSE;
	}

	rect->right = rCol1.left;

	return TRUE;
}

int MainDialog::GetPending(void)
{
	int num, count, c;

	num = ListView_GetItemCount( hList );

	for ( c = nCurItem, count = 0; c < num; c++ ) {
		LVITEM lvi;
		lvi.mask = LVIF_IMAGE;
		lvi.iItem = c;

		lvi.iSubItem = 0;
		ListView_GetItem( hList, &lvi );
		if ( lvi.iImage == -1 ) {
			count++;
		}
	}

	return count;
}

void MainDialog::UpdateControls(void)
{
	int count;

	count = GetPending();

	if ( m_ptrTaskbarList3 ) {
		TCHAR temp[256];
		int badge = min( 10, count );
		if ( bInDemux ) {
			badge = min( 10, count - 1 );
			if ( badge < 0 ) badge = 0;
		}
		if ( badge < 10 ) {
			_stprintf_s( temp, _countof(temp), _T("%d"), badge );
		}
		else {
			_tcscpy_s( temp, _countof(temp), _T("9+") );
		}
		HICON hIcon = ( badge > 0 ) ? LoadIcon( hInst, MAKEINTRESOURCE( IDI_ICON1 + badge - 1 ) ) : NULL;
		m_ptrTaskbarList3->SetOverlayIcon( m_hDlg, hIcon, temp );
	}

	EnableItem( IDC_BTN_START, ( ( count > 0 ) || ( bInDemux ) ) ? TRUE : FALSE );

	if ( bInDemux ) {
		SetItemText( IDC_BTN_START, _T("~ (&S)") );
	}
	else {
		SetItemText( IDC_BTN_START, _T("Jn (&S)") );
	}
}

BOOL MainDialog::Demux(void)
{
	int num, c;

	if ( nCurItem == -1 ) nCurItem = 0;

	num = ListView_GetItemCount( hList );

	for ( c = nCurItem; c < num; c++ ) {
		LVITEM lvi;
		lvi.mask = LVIF_IMAGE;
		lvi.iItem = c;

		lvi.iSubItem = 0;
		ListView_GetItem( hList, &lvi );
		if ( lvi.iImage == -1 ) {
			nCurItem = c;
			TCHAR path[_MAX_PATH];
			lvi.mask = LVIF_TEXT;
			lvi.pszText = path;
			lvi.cchTextMax = _countof(path);
			lvi.iSubItem = 1;
			ListView_GetItem( hList, &lvi );

			m_pDemux = new DemuxThread();

			if ( GetCheck( IDC_CHECK_ERRORSPLIT ) ) {
				m_pDemux->retrySplit = TRUE;
			}

			m_pDemux->Demux( lvi.pszText, m_hDlg );
			return TRUE;
		}
	}
	return FALSE;
}

BOOL MainDialog::Next( PROG_STATE curState )
{
	BOOL bRet = TRUE;
	LVITEM lvi;
	lvi.mask = LVIF_IMAGE;
	lvi.iItem = nCurItem;
	lvi.iSubItem = 0;
	lvi.iImage = GetImageIndex( curState );
	ListView_SetItem( hList, &lvi );

	m_pDemux->Exit();
	delete m_pDemux;
	m_pDemux = NULL;

	nCurItem++;
	if ( bInDemux && ( GetPending() > 0 ) ) {
		bRet = Demux();
		UpdateControls();
	}
	else {
		bInDemux = FALSE;
		UpdateControls();

		if ( m_ptrTaskbarList3 ) {
			m_ptrTaskbarList3->SetProgressState( m_hDlg, TBPF_NOPROGRESS );
		}

		FLASHWINFO fwi;
		fwi.cbSize = sizeof(FLASHWINFO);
		fwi.hwnd = m_hDlg;
		fwi.dwFlags = FLASHW_ALL | FLASHW_TIMERNOFG;
		fwi.uCount = 0;
		fwi.dwTimeout = 0;
		FlashWindowEx( &fwi );

		if ( GetCheck( IDC_CHECK_SUSPEND ) ) {
			SetSuspendState( FALSE, FALSE, FALSE );
		}
	}
	return bRet;
}

void MainDialog::CalcMinSize(void)
{
	RECT rect1, rectDlg;

	GetWindowRect( m_hDlg, &rectDlg );

	GetDlgItemRect( IDC_LIST, &rect1 );
	ptMinSize.x = rectDlg.right - rectDlg.left;
	ptMinSize.y = rectDlg.bottom - rectDlg.top - ( rect1.bottom - rect1.top );

	DWORD lvSize = ListView_ApproximateViewRect( hList, -1, -1, 1 );
	ptMinSize.y += HIWORD(lvSize);
}

BOOL MainDialog::IsValidExt( TCHAR *szPath )
{
	const TCHAR validExts[][6] = { _T(".ts"), _T(".m2ts") };
	TCHAR ext[_MAX_EXT];
	BOOL bRet = FALSE;

	_tsplitpath_s( szPath, NULL, 0, NULL, 0, NULL, 0, ext, _MAX_EXT );

	for ( int c = 0; c < _countof(validExts); c++ ) {
		if ( _tcsicmp( ext, validExts[c] ) == 0 ) {
			bRet = TRUE;
			break;
		}
	}
	return bRet;
}

void MainDialog::MainSetMin( POINT *pt )
{
	int height;
	LONG_PTR listStyle;

	listStyle = GetWindowLongPtr( hList, GWL_STYLE );
	if ( listStyle & WS_HSCROLL ) {
		height = GetSystemMetrics( SM_CYHSCROLL );
		pt->y += height;
	}
}

BOOL MainDialog::Cls_OnInitDialog(HWND hwnd, HWND hwndFocus, LPARAM lParam)
{
	hList = GetDlgItem( m_hDlg, IDC_LIST );
	InsertSystemMenu();
	InitControls();
	OleInitialize( NULL );
	pDropTarget = new CDropTarget( m_hDlg, &IsValidExt );
	RegisterDragDrop( m_hDlg, pDropTarget );
	CalcMinSize();

	return FALSE;
}

void MainDialog::Cls_OnDropFiles(HWND hwnd, HDROP hdrop)
{
	UINT uiSize;
	TCHAR szPath[_MAX_PATH];

	uiSize = DragQueryFile( hdrop, 0xFFFFFFFF, NULL, 0 );
	if ( uiSize > 0 ) {
		int badge = min( 10, uiSize );
		for ( UINT c = 0; c < uiSize; c++ ) {
			DragQueryFile( hdrop, c, szPath, _countof(szPath) );
			ListInsertItem( szPath );
		}
	}
	DragFinish( hdrop );
	UpdateControls();
}

void MainDialog::Cls_OnSysCommand(HWND hwnd, UINT cmd, int x, int y)
{
	switch( cmd & 0xfff0 )
	{
	case IDM_SYS_ABOUT:
		if ( m_aboutDlg ) {
			m_aboutDlg->SetFocus();
		}
		else {
			m_aboutDlg = new AboutDlg(IDD_ABOUTBOX);
			m_aboutDlg->DoModal(m_hDlg);
			delete m_aboutDlg;
			m_aboutDlg = NULL;
		}
		break;

	default:
		DefWindowProc(hwnd, WM_SYSCOMMAND, cmd, MAKELPARAM( x, y ) );
	}
}

void MainDialog::Cls_OnCommand(HWND hwnd, int id, HWND hwndCtl, UINT codeNotify)
{
	switch( id )
	{
	case IDC_BTN_START:
		if ( !bInDemux ) {
			bInDemux = TRUE;
			nCurItem = ListView_GetNextItem( hList, -1, LVNI_ALL | LVNI_SELECTED );
			Demux();
		}
		else {
			bInDemux = FALSE;
			m_pDemux->Exit();
		}
		UpdateControls();
		break;

	case IDOK:
	case IDCANCEL:
		if ( bInDemux ) {
			bInDemux = FALSE;
			m_pDemux->Exit();
		}
		DestroyWindow( m_hDlg );
		break;
	}
}

void MainDialog::Cls_OnDestroy(HWND hwnd)
{
	RevokeDragDrop( m_hDlg );
	if (pDropTarget != NULL) {
		pDropTarget->Release();
		//delete pDropTarget;	sv
		pDropTarget = NULL;
	}
	ImageList_Destroy( imageList );
	OleUninitialize();
	PostQuitMessage( 0 );
}

BOOL MainDialog::Cls_OnNotify( WPARAM wParam, LPARAM lParam )
{
	LPNMHDR lpnmhdr = (LPNMHDR)lParam;

	if ( lpnmhdr->hwndFrom == hList ) {

		LPNMLISTVIEW lplv = (LPNMLISTVIEW)lParam;

		if(lplv->hdr.code == NM_CUSTOMDRAW) {

			//MessageBox( m_hDlg, _T("Hit!"), _T("Test"), MB_OK );
			LPNMLVCUSTOMDRAW lplvcd = (LPNMLVCUSTOMDRAW)lParam;
			INT_PTR ret = ListCustom( lplvcd );

			if ( ret ) return ret;
		}
	}
	return FALSE;
}

void MainDialog::Cls_OnDemuxProgress( int progress )
{
	TCHAR szPath[_MAX_PATH];

	_stprintf_s( szPath, _countof(szPath), _T("%d%%"), progress );
	ListView_SetItemText( hList, nCurItem, 2, szPath );
	if ( m_ptrTaskbarList3 ) {
		m_ptrTaskbarList3->SetProgressValue( m_hDlg, progress, 100 );
	}
}

BOOL MainDialog::FillSolidRect( HDC hdc, RECT *rect, COLORREF color )
{
	if ( SetBkColor( hdc, color ) == CLR_INVALID ) {
		return FALSE;
	}
	return ExtTextOut( hdc, 0, 0, ETO_OPAQUE, rect, NULL, 0, NULL );
}

void MainDialog::Cls_OnDemuxSplit( int errorCount )
{
	LVITEM lvi;
	TCHAR count[20];

	_stprintf_s( count, _countof(count), _T("%d"), errorCount );

	lvi.mask = LVIF_TEXT;
	lvi.iItem = nCurItem;
	lvi.iSubItem = 0;
	lvi.pszText = count;
	ListView_SetItem( hList, &lvi );
}
