#ifndef _DVHS_PCM_H_
#define _DVHS_PCM_H_

#include "common.h"

namespace dvhs
{
	class wavconverter
	{
	private:
		int stage;
		unsigned char tmp;
		unsigned long packet_size, pos_in_packet;
		unsigned char *wp;
		u_int64_t skip_bytes;

	public:
		unsigned char wav_buf[200], wav_tmp[4];
		bool skip_set;
		unsigned long output_bytes;
		wavconverter(void) { reset(); };

		void set_skip_bytes( long long bytes_diff )
		{
			if ( bytes_diff > 0 ) {
				skip_bytes = bytes_diff;
			}
		}

		int convert1(unsigned char t)
		{
			int toadd = 0;
			static int tblCnv[16]={
				0,8,4,12,2,10,6,14,1,9,5,13,3,11,7,15
			};

			switch(stage)
			{
			case 0:
				packet_size = t << 8;
				stage++;
				break;
			case 1:
				packet_size |= t;
				stage++;
				break;
			case 2:
				stage++;
				break;
			case 3:
				stage++;
				pos_in_packet = 0;
				break;
			case 4:
				tmp = t;
				pos_in_packet++;
				stage++;
				break;
			case 5:
				//*wp++ = tblCnv[ tmp & 0x0f ] | ( tblCnv[ t >> 4 ] << 4 );
				wav_tmp[0] = tblCnv[ tmp & 0x0f ] | ( tblCnv[ t >> 4 ] << 4 );
				tmp = t;
				//toadd++;
				pos_in_packet++;
				stage++;
				break;
			case 6:
				//*wp++ = tblCnv[ tmp & 0x0f ] | ( tblCnv[ t >> 4 ] << 4 );
				wav_tmp[1] = tblCnv[ tmp & 0x0f ] | ( tblCnv[ t >> 4 ] << 4 );
				tmp = t;
				//toadd++;
				pos_in_packet++;
				stage++;
				break;
			case 7:
				//*wp++ = tblCnv[ t >> 4 ] | ( tblCnv[ t & 0x0f ] << 4 );
				wav_tmp[2] = tblCnv[ t >> 4 ] | ( tblCnv[ t & 0x0f ] << 4 );
				//toadd++;
				pos_in_packet++;
				stage++;
				break;
			case 8:
				//*wp++ = tblCnv[ t >> 4 ] | ( tblCnv[ t & 0x0f ] << 4 );
				wav_tmp[3] = tblCnv[ t >> 4 ] | ( tblCnv[ t & 0x0f ] << 4 );

				if ( skip_bytes > 0 ) {
					skip_bytes -= 4;
				}
				else {
					toadd+=4;
					memcpy( wp, wav_tmp, 4 );
					wp += 4;
				}
				pos_in_packet++;
				if ( pos_in_packet >= packet_size ) {
					stage = 0;
				}
				else {
					stage = 4;
				}
				break;
			}
			return toadd;
		}

		int convert(const char* p,int l)
		{
			int c, r = 0;

			wp = wav_buf;
			for ( c = 0; c < l; c++ ) {
				r += convert1( p[c] );
			}

			output_bytes += r;

			return r;
		}
		int flush(void)
		{
			unsigned long z = ( ( packet_size / 5 ) - ( pos_in_packet / 5 ) ) * 4;
			output_bytes += z;

			return z;
		}
		unsigned long get_size(void) {return output_bytes;}
		void reset(void)
		{
			packet_size = 0;
			stage = 0;
			tmp = 0;
			output_bytes = 0;
			skip_bytes = 0;
			skip_set = false;
			pos_in_packet = 0;
			wp = NULL;
			memset( wav_buf, 0, sizeof(wav_buf) );
			memset( wav_tmp, 0, sizeof(wav_tmp) );
		}
	};
}

#endif//_DVHS_PCM_H_
