/*

 Copyright (C) 2009 Anton Burdinuk

 clark15b@gmail.com

*/


#include "stdafx.h"
#include "ts.h"
#include "misc.h"

// TODO: join TS

// 1) M2TS timecode (0-1073741823)
/*
The extra 4-byte header is composed of two fields. The upper 2 bits are the copy_permission_indicator and the lower 30 bits are the arrival_time_stamp. The arrival_time_stamp is equal to the lower 30 bits of the 27 MHz STC at the 0x47 byte of the Transport packet. In a packet that contains a PCR, the PCR will be a few ticks later than the arrival_time_stamp. The exact difference between the arrival_time_stamp and the PCR (and the number of bits between them) indicates the intended fixed bitrate of the variable rate Transport Stream.

The primary function is to allow the variable rate Transport Stream to be converted to a fixed rate stream before decoding (since only fixed rate Transport Streams fit into the T-STD buffering model).

It doesn't really help for random access. 30 bits at 27 MHz only represents 39.77 seconds.
*/

// 2) TS Continuity counter (0-15..0-15..)
// 3) PES PTS/DTS

namespace ts
{
	void get_prefix_name_by_filename(const tstring& s,tstring& name, u_int32_t error_counter);
#ifdef _WIN32
	void my_strptime(const char* s,tm* t);
#endif
}

ts::stream::~stream(void)
{
	if(timecodes)
		fclose(timecodes);
}


void ts::get_prefix_name_by_filename(const tstring& s,tstring& name, u_int32_t error_counter)
{
#ifdef _WIN32
	TCHAR path[_MAX_PATH], drive[_MAX_DRIVE], dir[_MAX_DIR], fname[_MAX_FNAME], ext[_MAX_EXT];

	_tsplitpath_s( s.c_str(), drive, _MAX_DRIVE, dir, _MAX_DIR, fname, _MAX_FNAME, ext, _MAX_EXT );
	_tmakepath_s( path, _MAX_PATH, drive, dir, fname, NULL );
	name = path;

	if ( error_counter > 0 ) {
		_stprintf_s( path, _MAX_PATH, _T("_%d"), error_counter );
		name += path;
	}
#else
	int ll=s.length();
	const TCHAR* p=s.c_str();

	while(ll>0)
	{
		if(p[ll-1]==_T('/') || p[ll-1]==_T('\\'))
			break;
		ll--;
	}

	p+=ll;

	int cn=0;

	const TCHAR* pp=_tcschr(p,_T('.'));

	if(pp)
		name.assign(p,pp-p);
#endif
}


ts::file::~file(void)
{
	close();
}

bool ts::file::open(int mode,const TCHAR* fmt,...)
{
	filename.clear();

	TCHAR name[512];

	va_list ap;
	va_start(ap,fmt);
	_vstprintf_s(name,512,fmt,ap);
	va_end(ap);

	int flags=0;

	switch(mode)
	{
	case in:
		flags=O_LARGEFILE|O_BINARY|O_RDONLY;
		break;
	case out:
		flags=O_CREAT|O_TRUNC|O_LARGEFILE|O_BINARY|O_WRONLY;
		break;
	}

	//fd=::_topen(name,flags,0644);

	//if(fd!=-1)
	if(::_tsopen_s(&fd,name,flags,_SH_DENYWR,_S_IREAD|_S_IWRITE)==0)
	{
		filename=name;
		len=offset=0;
		return true;
	}

	fprintf(stderr,"can`t open file %s\n",name);

	return false;
}


void ts::file::close(void)
{
	if(fd!=-1)
	{
		flush();
		::_close(fd);
		fd=-1;
	}
	len=0;
	offset=0;
}

int ts::file::flush(void)
{
	int l=0;

	while(l<len)
	{
		int n=::_write(fd,buf+l,len-l);
		if(!n || n==-1)
			break;
		l+=n;
	}

	len=0;

	return l;
}

int ts::file::write(const char* p,int l)
{
	int rc=l;

	while(l>0)
	{
		if(len>=max_buf_len)
			flush();

		int n=max_buf_len-len;

		if(n>l)
			n=l;

		memcpy(buf+len,p,n);
		len+=n;

		p+=n;
		l-=n;
	}

	return rc;
}
int ts::file::read(char* p,int l)
{
	const char* tmp=p;

	while(l>0)
	{
		int n=len-offset;

		if(n>0)
		{
			if(n>l)
				n=l;

			memcpy(p,buf+offset,n);
			p+=n;
			offset+=n;
			l-=n;
		}else
		{
			int m=::_read(fd,buf,max_buf_len);
			if(m==-1 || !m)
				break;
			len=m;
			offset=0;
		}
	}

	return p-tmp;
}
int ts::file::write_wav_header(void)
{
	unsigned char wav[] = { 0x52, 0x49, 0x46, 0x46, 0x00, 0x00, 0x00, 0x00, 0x57, 0x41, 0x56, 0x45, 0x66, 0x6D, 0x74, 0x20,
							0x10, 0x00, 0x00, 0x00, 0x01, 0x00, 0x02, 0x00, 0x80, 0xBB, 0x00, 0x00, 0x00, 0xEE, 0x02, 0x00,
							0x04, 0x00, 0x10, 0x00, 0x64, 0x61, 0x74, 0x61, 0x00, 0x00, 0x00, 0x00 };
	wav_header_size = sizeof(wav);

	return write( (const char*)wav, sizeof(wav) );
}


bool ts::demuxer::validate_type(u_int8_t type)
{
	if(av_only)
		return strchr("\x01\x02\x80\x1b\xea\x81\x06\x83\x03\x04\x82\x86\x8a",type)?true:false;

	return true;
}

int ts::demuxer::get_stream_type(u_int8_t type)
{
	switch(type)
	{
	case 0x01:
	case 0x02:
		return stream_type::mpeg2_video;
	case 0x80:
		return hdmv?stream_type::lpcm_audio:stream_type::mpeg2_video;
	case 0x1b:
		return stream_type::h264_video;
	case 0xea:
		return stream_type::vc1_video;
	case 0x81:
	case 0x06:
	case 0x83:
		return dvhs?stream_type::dvhs_pcm_audio:stream_type::ac3_audio;
	case 0x03:
	case 0x04:
		return stream_type::mpeg2_audio;
	case 0x82:
	case 0x86:
	case 0x8a:
		return stream_type::dts_audio;
	}
	return stream_type::data;
}

const TCHAR* ts::demuxer::get_stream_ext(u_int8_t type_id)
{
	static const TCHAR* list[9]= { _T("sup"), _T("m2v"), _T("264"), _T("vc1"), _T("ac3"), _T("m2a"), _T("pcm"), _T("dts"), _T("wav") };

	if(type_id<0 || type_id>=9)
		type_id=0;

	return list[type_id];
}

u_int64_t ts::demuxer::decode_pts(const char* ptr)
{
	const unsigned char* p=(const unsigned char*)ptr;

	u_int64_t pts=((((u_int64_t)p[0])&0xe)<<29);
	pts|=((p[1]&0xff)<<22);
	pts|=((p[2]&0xfe)<<14);
	pts|=((p[3]&0xff)<<7);
	pts|=((p[4]&0xfe)>>1);

	return pts;
}

int ts::demuxer::demux_ts_packet(const char* ptr)
{
	u_int32_t timecode=0;
	if(hdmv)
	{
		timecode=to_int32(ptr)&0x3fffffff;
		ptr+=4;
	}

	const char* end_ptr=ptr+188;

	if(ptr[0]!=0x47)			// ts sync byte
		return -1;

	u_int16_t pid=to_int(ptr+1);
	u_int8_t flags=to_byte(ptr+3);

	bool transport_error=(pid&0x8000)?true:false;
	bool payload_unit_start_indicator=(pid&0x4000)?true:false;
	bool adaptation_field_exist=(flags&0x20)?true:false;
	bool payload_data_exist=(flags&0x10)?true:false;
	u_int8_t continuity_counter=(flags&0x0f);
	pid&=0x1fff;

	if(transport_error)
		return -2;

	if(pid==0x1fff || !payload_data_exist)
		return 0;

	ptr+=4;

	// skip adaptation field
	if(adaptation_field_exist)
	{
		ptr+=to_byte(ptr)+1;
		if(ptr>=end_ptr)
			return -3;
	}

	if(dump==1)
		printf("%.4x: [%c%c%c%c] %u.%i\n",
			pid,
			transport_error?'e':'-',
			payload_data_exist?'p':'-',
			payload_unit_start_indicator?'s':'-',
			adaptation_field_exist?'a':'-',
			timecode,
			continuity_counter
			);


	stream& s=streams[pid];

	if ( s.continuity_counter == 0xff ) {
		s.continuity_counter = continuity_counter;
	}
	else {
		u_int8_t prev_counter = s.continuity_counter;
		prev_counter++;
		if ( prev_counter > 0x0f ) {
			prev_counter = 0;
		}
		if ( prev_counter != continuity_counter ) {
			int stype = get_stream_type(streams[pid].type);
			if ( ( !ignore_m3a && ( stype == stream_type::mpeg2_audio ) ) || ( stype == stream_type::mpeg2_video ) || ( stype == stream_type::dvhs_pcm_audio ) ) {
				return -30;
			}
		}
		else {
			s.continuity_counter = continuity_counter;
		}
	}

	if(!pid || (s.channel!=0xffff && s.type==0xff))
	{
		// PSI

		if(payload_unit_start_indicator)
		{
			// begin of PSI table

			ptr++;

			if(ptr>=end_ptr)
				return -4;

			if(*ptr!=0x00 && *ptr!=0x02)
				return 0;

			if(end_ptr-ptr<3)
				return -5;

			u_int16_t l=to_int(ptr+1);

			if((l&0x3000)!=0x3000)
				return -6;

			l&=0x0fff;

			ptr+=3;

			int len=end_ptr-ptr;

			if(l>len)
			{
				if(l>ts::table::max_buf_len)
					return -7;

				s.psi.reset();

				memcpy(s.psi.buf,ptr,len);
				s.psi.offset+=len;
				s.psi.len=l;

				return 0;
			}else
				end_ptr=ptr+l;
		}else
		{
			// next part of PSI
			if(!s.psi.offset)
				return -8;

			int len=end_ptr-ptr;

			if(len>ts::table::max_buf_len-s.psi.offset)
				return -9;

			memcpy(s.psi.buf+s.psi.offset,ptr,len);
			s.psi.offset+=len;

			if(s.psi.offset<s.psi.len)
				return 0;
			else
			{
				ptr=s.psi.buf;
				end_ptr=ptr+s.psi.len;
			}
		}

		if(!pid)
		{
			// PAT

			ptr+=5;

			if(ptr>=end_ptr)
				return -10;

			int len=end_ptr-ptr-4;

			if(len<0 || len%4)
				return -11;

			int n=len/4;

			for(int i=0;i<n;i++,ptr+=4)
			{
				u_int16_t channel=to_int(ptr);
				u_int16_t pid=to_int(ptr+2);

				if((pid&0xe000)!=0xe000)
					return -12;

				pid&=0x1fff;

				if(!demuxer::channel || demuxer::channel==channel)
				{
					stream& ss=streams[pid];
					ss.channel=channel;
					ss.type=0xff;
				}
			}
		}else
		{
			// PMT

			ptr+=7;

			if(ptr>=end_ptr)
				return -13;

			u_int16_t info_len=to_int(ptr)&0x0fff;

			ptr+=info_len+2;
			end_ptr-=4;

			if(ptr>=end_ptr)
				return -14;

			while(ptr<end_ptr)
			{
				if(end_ptr-ptr<5)
					return -15;

				u_int8_t type=to_byte(ptr);
				u_int16_t pid=to_int(ptr+1);

				if((pid&0xe000)!=0xe000)
					return -16;

				pid&=0x1fff;

				info_len=to_int(ptr+3)&0x0fff;

				ptr+=5+info_len;

				// ignore unknown streams
				if(validate_type(type))
				{
					stream& ss=streams[pid];

					if(ss.channel!=s.channel || ss.type!=type)
					{
						ss.channel=s.channel;
						ss.type=type;
						ss.id=++s.id;

						if(!parse_only && !ss.file.is_opened() && ( ( get_stream_type(ss.type) != stream_type::mpeg2_audio ) || ( !ignore_m3a && ( get_stream_type(ss.type) == stream_type::mpeg2_audio ) ) ) )
						{
							if(dst.length())
								ss.file.open(file::out,_T("%s%c%strack_%i.%s"),dst.c_str(),os_slash,prefix.c_str(),pid,get_stream_ext(get_stream_type(ss.type)));
							else if ( !not_add_track )
								ss.file.open(file::out,_T("%strack_%i.%s"),prefix.c_str(),pid,get_stream_ext(get_stream_type(ss.type)));
							else
								ss.file.open(file::out,_T("%s%s"),prefix.c_str(),get_stream_ext(get_stream_type(ss.type)));

							if ( get_stream_type(ss.type) == stream_type::dvhs_pcm_audio ) {
								ss.file.write_wav_header();
							}
						}
					}
				}
			}

			if(ptr!=end_ptr)
				return -18;
		}
	}else
	{
		if(s.type!=0xff)
		{
			// PES

			if(payload_unit_start_indicator)
			{
				s.psi.reset();
				s.psi.len=9;
			}

			while(s.psi.offset<s.psi.len)
			{
				int len=end_ptr-ptr;

				if(len<=0)
					return 0;

				int n=s.psi.len-s.psi.offset;

				if(len>n)
					len=n;

				memcpy(s.psi.buf+s.psi.offset,ptr,len);
				s.psi.offset+=len;

				ptr+=len;

				if(s.psi.len==9)
					s.psi.len+=to_byte(s.psi.buf+8);
			}

			if(s.psi.len)
			{
				if(memcmp(s.psi.buf,"\x00\x00\x01",3))
					return -19;

				s.stream_id=to_byte(s.psi.buf+3);

				if ( s.stream_id == 0xe0 ) {
					const unsigned char group_start_code[]	= { 0x00, 0x00, 0x01, 0xb8 };
					const unsigned char picture_start_code[] = { 0x00, 0x00, 0x01, 0x00 };
					const char *p;
					bool group_found = false;
					bool picture_found = false;
					char picture_type;

					for ( p = ptr; p < ( end_ptr - 10 ); p++ ) {
						if ( memcmp( p, group_start_code, 4 ) == 0 ) {
							group_found = true;
							break;
						}
					}
					if ( group_found ) {
						for ( ; p < ( end_ptr - 10 ); p++ ) {
							if ( memcmp( p, picture_start_code, 4 ) == 0 ) {
								picture_found = true;
								break;
							}
						}
					}
					if ( picture_found ) {
						picture_type = ( *( p + 5 ) >> 3 ) & 7;
						// I frame ?
						if ( picture_type = 1 ) {
							gop_counter++;
						}
					}
				}

//				if ( gop_counter == 0 ) return 0;

				if ( ( s.stream_id != 0 ) && ( gop_counter > 0 ) ) {
					s.stream_counter++;
				}

//				if ( s.stream_counter == 0 ) return 0;

				u_int8_t flags=to_byte(s.psi.buf+7);

				s.frame_num++;

				switch(flags&0xc0)
				{
				case 0x80:			// PTS only
					{
						u_int64_t pts=decode_pts(s.psi.buf+9);

						if(dump==2)
							printf("%.4x: %llu\n",pid,pts);
						else if(dump==3)
							printf("%.4x: track=%.4x.%.2i, type=%.2x, stream=%.2x, pts=%llums\n",pid,s.channel,s.id,s.type,s.stream_id,pts/90);

						if(s.dts>0 && pts>s.dts)
							s.frame_length=pts-s.dts;
						s.dts=pts;

						if(pts>s.last_pts)
							s.last_pts=pts;

						if((!s.first_pts)||(s.stream_counter<=1)) {
							s.first_pts=pts;
							if ( s.stream_id == 0xe0 ) {
								base_pts = pts;
							}
						}
					}
					break;
				case 0xc0:			// PTS,DTS
					{
						u_int64_t pts=decode_pts(s.psi.buf+9);
						u_int64_t dts=decode_pts(s.psi.buf+14);

						if(dump==2)
							printf("%.4x: %llu %llu\n",pid,pts,dts);
						else if(dump==3)
							printf("%.4x: track=%.4x.%.2i, type=%.2x, stream=%.2x, pts=%llums, dts=%llums\n",pid,s.channel,s.id,s.type,s.stream_id,pts/90,dts/90);

						if(s.dts>0 && dts>s.dts)
							s.frame_length=dts-s.dts;
						s.dts=dts;

						if(pts>s.last_pts)
							s.last_pts=pts;

						if((!s.first_dts)||(s.stream_counter<=1))
							s.first_dts=dts;

						if((!s.first_pts)||(s.stream_counter<=1)) {
							s.first_pts=pts;
							if ( s.stream_id == 0xe0 ) {
								base_pts = pts;
							}
						}
					}
					break;
				}

				if(pes_output && s.file.is_opened())
					s.file.write(s.psi.buf,s.psi.len);

				s.psi.reset();
			}

			if(s.frame_num)
			{
				int len=end_ptr-ptr;

				if(es_parse)
				{
					switch(s.type)
					{
					case 0x1b:
						s.frame_num_h264.parse(ptr,len);
						break;
					case 0x06:
					case 0x81:
					case 0x83:
						s.frame_num_ac3.parse(ptr,len);
						break;
					}
				}

				if(s.file.is_opened()&&(s.stream_counter > 0)) {
					if ( get_stream_type(s.type) == stream_type::dvhs_pcm_audio ) {
						if ( ( adjust_wav ) && ( !wavconv.skip_set ) ) {
							long long bytes_diff = ( ( (long long)s.first_pts - (long long)base_pts ) * 48000 / 90000 ) * 4;
							if ( bytes_diff < 0 ) {
								wavconv.set_skip_bytes( -bytes_diff );
							}
							else if ( bytes_diff > 0 ) {
								const char padding[4] = {0};
								for ( long long pad_counter = 0; pad_counter < ( bytes_diff / 4 ); pad_counter++ ) {
									s.file.write( padding, 4 );
								}
								wavconv.output_bytes += (unsigned long)bytes_diff;
							}
							wavconv.skip_set = true;
						}
						s.file.write( (const char*)wavconv.wav_buf, wavconv.convert( ptr, len ) );
					}
					else {
						s.file.write(ptr,len);
					}
				}
			}
		}
	}

	return 0;
}

void ts::demuxer::show(void)
{
	u_int64_t beg_pts=0,end_pts=0;

	for(std::map<u_int16_t,ts::stream>::const_iterator i=streams.begin();i!=streams.end();++i)
	{
		const ts::stream& s=i->second;

		if(s.type!=0xff)
		{
			if(s.first_pts<beg_pts || !beg_pts)
				beg_pts=s.first_pts;

			u_int64_t n=s.last_pts+s.frame_length;
			if(n>end_pts || !end_pts)
				end_pts=n;
		}
	}

	for(std::map<u_int16_t,ts::stream>::const_iterator i=streams.begin();i!=streams.end();++i)
	{
		u_int16_t pid=i->first;
		const ts::stream& s=i->second;

		if(s.type!=0xff)
		{
			u_int64_t end=s.last_pts+s.frame_length;
			u_int64_t len=end-s.first_pts;

			fprintf(stderr,"pid=%i (0x%.4x), ch=%i, id=%.i, type=0x%.2x (%s), stream=0x%.2x",
				pid,pid,s.channel,s.id,s.type,get_stream_ext(get_stream_type(s.type)),s.stream_id);

			if(s.frame_length>0)
				fprintf(stderr,", fps=%.2f",90000./(double)s.frame_length);

			if(len>0)
				fprintf(stderr,", len=%llums",len/90);


			if(s.frame_num>0)
				fprintf(stderr,", fn=%llu",s.frame_num);

			u_int64_t esfn=s.get_es_frame_num();

			if(esfn>0)
				fprintf(stderr,", esfn=%llu",esfn);

			if(s.first_pts>beg_pts)
			{
				u_int64_t n=(s.first_pts-beg_pts)/90;

				fprintf(stderr,", head=+%llums",n);
			}

			if(end<end_pts)
			{
				u_int64_t n=(end_pts-end)/90;

				fprintf(stderr,", tail=-%llums",n);
			}

			fprintf(stderr,"\n");
		}
	}
}

int ts::demuxer::demux_file(const TCHAR* name)
{
	int prg, prevPrg = -1;
	__int64 curpos = 0, filesize;
	u_int64_t first_ts_v = 0;
	long long dts_diff_a = 0;
	FILE *logFile = NULL;
	TCHAR logName[_MAX_PATH];
	fpos_t fileSize;
	int returnCode = 0;
	u_int32_t error_counter = 0;
	u_int32_t filename_index = 0;
	bool retry_with_split = false;
	bool delete_short_file = false;

	prefix.clear();

	char buf[192];

	int buf_len=0;

	ts::file file;

	if(!file.open(file::in,_T("%s"),name))
	{
		_ftprintf(stderr,_T("can`t open file %s\n"),name);
		return -1;
	}
	filesize = file.size();

SPLIT_RETRY:
	get_prefix_name_by_filename(name,prefix,filename_index);
	if(prefix.length())
		prefix+=_T('.');
	retry_with_split = false;

	for(u_int64_t pn=1;loop_continue;pn++)
	{
		if(buf_len)
		{
			if(file.read(buf,buf_len)!=buf_len)
				break;
		}else
		{
			if(file.read(buf,188)!=188)
				break;
			if(buf[0]==0x47 && buf[4]!=0x47)
			{
				buf_len=188;
				_ftprintf(stderr,_T("TS stream detected in %s (packet length=%i)\n"),name,buf_len);
				hdmv=false;
			}else if(buf[0]!=0x47 && buf[4]==0x47)
			{
				if(file.read(buf+188,4)!=4)
					break;
				buf_len=192;
				_ftprintf(stderr,_T("M2TS stream detected in %s (packet length=%i)\n"),name,buf_len);
				hdmv=true;
			}else
			{
				_ftprintf(stderr,_T("unknown stream type in %s\n"),name);
				returnCode = -1;
				loop_continue = false;
				break;
			}
		}

		curpos += buf_len;
		prg = (int)( curpos * 100 / filesize );
		if ( prevPrg != prg ) {
			if ( target_hwnd ) {
				PostMessage( target_hwnd, WM_DEMUXPROGRESS, prg, 0 );
				prevPrg = prg;
			}
		}

		int n;
		if((n=demux_ts_packet(buf)))
		{
			_ftprintf(stderr,_T("%s: invalid packet %llu (%i)\n"),name,pn,n);
			returnCode = -1;
			
			if ( split_with_error ) {
				error_counter++;
				retry_with_split = true;
			}
			else {
				loop_continue = false;
			}
			break;
		}
		Sleep(0);
	}

	_stprintf_s( logName, _T("%slog"), prefix.c_str() );
	_tfopen_s( &logFile, logName, _T("wt") );

	for(std::map<u_int16_t,ts::stream>::const_iterator i=streams.begin();i!=streams.end();++i) {
		ts::stream &s = (ts::stream&)i->second;
		if ( get_stream_type(s.type) == stream_type::mpeg2_video ) {
			if ( s.first_dts ) first_ts_v = s.first_dts;
			if ( s.first_pts ) first_ts_v = s.first_pts;

			for(std::map<u_int16_t,ts::stream>::const_iterator j=streams.begin();j!=streams.end();++j)
			{
				ts::stream &s1 = (ts::stream&)j->second;
				if ( get_stream_type(s1.type) >= stream_type::ac3_audio ) {

					if ( ( adjust_wav ) && ( get_stream_type(s1.type) == stream_type::dvhs_pcm_audio ) ) continue;
					if ( ignore_m3a && ( get_stream_type(s1.type) == stream_type::mpeg2_audio ) ) continue;

					dts_diff_a = (long long)s1.first_pts - (long long)first_ts_v;

					double diff = dts_diff_a / 90.0;

					if ( logFile ) {
						_ftprintf( logFile, _T("%s : %.2fms\n"), get_stream_ext(get_stream_type(s1.type)), diff );
					}
				}
			}
			break;
		}
	}

	for(std::map<u_int16_t,ts::stream>::const_iterator i=streams.begin();i!=streams.end();++i) {
		ts::stream &s = (ts::stream&)i->second;
		if ( get_stream_type(s.type) == stream_type::dvhs_pcm_audio ) {
			int z = wavconv.flush();
			char *zero = new char[z]();	// initializing with 0
			s.file.flush();
			s.file.write( zero, z );
			s.file.flush();
			delete[] zero;

			unsigned long num = wavconv.get_size();
			unsigned char t[4];
			t[0] = num & 0xff;
			t[1] = ( num >>  8 ) & 0xff;
			t[2] = ( num >> 16 ) & 0xff;
			t[3] = ( num >> 24 ) & 0xff;
			s.file.write_direct( 0x28, (char *)t, 4 );
			num += ( 44 - 8 );
			t[0] = num & 0xff;
			t[1] = ( num >>  8 ) & 0xff;
			t[2] = ( num >> 16 ) & 0xff;
			t[3] = ( num >> 24 ) & 0xff;
			s.file.write_direct( 4, (char*)t, 4 );

			if ( s.file.size() < ( 192000 + s.file.get_wav_header_size() ) ) {
				delete_short_file = true;
			}
		}
		s.file.close();
	}

	if ( logFile ) {
		fseek( logFile, 0, SEEK_END );
		fgetpos( logFile, &fileSize );
		fclose( logFile );
		if ( fileSize == 0 ) _tunlink( logName );
	}

	if ( delete_short_file ) {
		for(std::map<u_int16_t,ts::stream>::const_iterator i=streams.begin();i!=streams.end();++i) {
			ts::stream &s = (ts::stream&)i->second;
			if ( !s.file.filename.empty() ) {
				_tunlink( s.file.filename.c_str() );
			}
		}
		_tunlink( logName );
	}

	if ( retry_with_split ) {
		gop_counter = 0;
		base_pts = 0;
		streams.clear();
		wavconv.reset();

		if ( !delete_short_file ) {
			filename_index++;
		}
		delete_short_file = false;

		if ( target_hwnd ) {
			PostMessage( target_hwnd, WM_DEMUXSPLIT, (WPARAM)error_counter, 0 );
		}

		goto SPLIT_RETRY;
	}

	if ( target_hwnd ) {
		PostMessage( target_hwnd, loop_continue ? WM_DEMUXFINISHED : WM_DEMUXABORTED, (WPARAM)error_counter, 0 );
	}

	return returnCode;
}

#ifdef _WIN32
void ts::my_strptime(const char* s,tm* t)
{
	memset((char*)t,0,sizeof(tm));
	sscanf_s(s,"%d-%d-%d %d:%d:%d",&t->tm_year,&t->tm_mon,&t->tm_mday,&t->tm_hour,&t->tm_min,&t->tm_sec);
	t->tm_year-=1900;
	t->tm_mon-=1;
	t->tm_isdst=1;
}
#endif

int ts::demuxer::gen_timecodes(const std::string& datetime)
{
	u_int64_t beg_pts=0;
	u_int64_t end_pts=0;

	for(std::map<u_int16_t,stream>::iterator i=streams.begin();i!=streams.end();++i)
	{
		u_int16_t pid=i->first;
		ts::stream& s=i->second;

		if(s.type!=0xff)
		{
			if(!s.timecodes && s.file.filename.length())
			{
				std::string::size_type n=s.file.filename.find_last_of(_T('.'));
				if(n!=std::string::npos)
				{
					tstring filename=s.file.filename.substr(0,n);
					filename+=_T(".tmc");

					//s.timecodes=_tfopen(filename.c_str(),_T("w"));

					//if(s.timecodes)
					if(!_tfopen_s(&s.timecodes,filename.c_str(),_T("w")))
						fprintf(s.timecodes,"# timecode format v2\n");
				}
			}
			if(s.timecodes)
			{
				if(s.first_pts<beg_pts || !beg_pts)
					beg_pts=s.first_pts;

				u_int64_t len=s.last_pts+s.frame_length;

				if(len>end_pts || !end_pts)
					end_pts=len;
			}
		}
	}


	for(std::map<u_int16_t,stream>::iterator i=streams.begin();i!=streams.end();++i)
	{
		u_int16_t pid=i->first;
		ts::stream& s=i->second;

		if(s.timecodes)
		{
			u_int64_t esfn=s.get_es_frame_num();

			u_int64_t frame_num=esfn?esfn:s.frame_num;

#ifdef OLD_TIMECODES
			write_timecodes(s.timecodes,base_pts+(s.first_pts-beg_pts),base_pts+(s.last_pts+s.frame_length-beg_pts),frame_num,s.frame_length);
#else
			switch(get_stream_type(s.type))
			{
			case stream_type::ac3_audio:
			case stream_type::mpeg2_audio:
			case stream_type::lpcm_audio:
				write_timecodes2(s.timecodes,base_pts+(s.first_pts-beg_pts),base_pts+(s.last_pts+s.frame_length-beg_pts),frame_num,s.frame_length);
				break;
			default:
				write_timecodes(s.timecodes,base_pts+(s.first_pts-beg_pts),base_pts+(s.last_pts+s.frame_length-beg_pts),frame_num,s.frame_length);
				break;
			}
#endif
		}
	}

	if(!subs && datetime.length())
	{
		TCHAR path[512];
		_stprintf_s(path,512,_T("%s%ctimecodes.srt"),dst.c_str(),os_slash);

		//subs=_tfopen(path,_T("w"));

		//if(subs)
		if(!_tfopen_s(&subs,path,_T("w")))
			subs_filename=path;
	}

	u_int64_t length=end_pts-beg_pts;

	if(subs && datetime.length())
	{
		u_int64_t cur=base_pts/90;

		u_int64_t num=(length/90)/1000;

		time_t timecode=0;

		tm t;
#ifdef _WIN32
		my_strptime(datetime.c_str(),&t);
#else
		strptime(datetime.c_str(),"%Y-%m-%d %H:%M:%S",&t);
#endif
		timecode=mktime(&t);

		for(u_int32_t i=0;i<num;i++)
		{
			std::string start=ts::timecode_to_time(cur);

			if(i<num-1)
				cur+=1000;
			else
				cur=(base_pts+length)/90;

			std::string end=ts::timecode_to_time(cur-1);

			start[8]=',';
			end[8]=',';

#ifdef _WIN32
			localtime_s(&t,&timecode);
#else
			localtime_r(&timecode,&t);
#endif

			char timecode_s[64];
			strftime(timecode_s,sizeof(timecode_s),"%Y-%m-%d %H:%M:%S",&t);

			fprintf(subs,"%u\n%s --> %s\n%s\n\n",++subs_num,start.c_str(),end.c_str(),timecode_s);

			timecode+=1;
		}
	}


	base_pts+=length;

#ifndef OLD_TIMECODES
	// align
	if(base_pts%90)
		base_pts=base_pts/90*90+90;
#endif

	return 0;
}

void ts::demuxer::write_timecodes(FILE* fp,u_int64_t first_pts,u_int64_t last_pts,u_int64_t frame_num,u_int64_t frame_len)
{
	u_int64_t len=last_pts-first_pts;

	double c=(double)len/(double)frame_num;

	double m=0;
	u_int32_t n=0;

	for(u_int32_t i=0;i<frame_num;i++)
	{
		fprintf(fp,"%llu\n",(first_pts+n)/90);

		m+=c;
		n+=(u_int32_t)c;

		if(m-n>=1)
			n+=1;
	}
}

#ifndef OLD_TIMECODES
// for audio
void ts::demuxer::write_timecodes2(FILE* fp,u_int64_t first_pts,u_int64_t last_pts,u_int64_t frame_num,u_int64_t frame_len)
{
	u_int64_t len=last_pts-first_pts;

	u_int64_t len2=frame_num*frame_len;

	if(len2>len || frame_len%90)
		write_timecodes(fp,first_pts,last_pts,frame_num,frame_len);
	else
	{
		u_int64_t frame_len_ms=frame_len/90;
		u_int64_t timecode_ms=base_pts/90;

		for(u_int64_t i=0;i<frame_num;i++)
		{
			fprintf(fp,"%llu\n",timecode_ms);
			timecode_ms+=frame_len_ms;
		}
	}
}
#endif

const char* ts::timecode_to_time(u_int64_t timecode)
{
	static char buf[128];

	int msec=timecode%1000;
	timecode/=1000;

	int sec=timecode%60;
	timecode/=60;

	int min=timecode%60;
	timecode/=60;

	sprintf_s(buf,128,"%.2i:%.2i:%.2i.%.3i",(int)timecode,min,sec,msec);

	return buf;
}
